#include "pitches.h"
#include <UTFTGLUE.h>              //use GLUE class and constructor
#include <Adafruit_MPU6050.h>

Adafruit_MPU6050 mpu;

UTFTGLUE myGLCD(0, A2, A1, A3, 13, A0); 

#define BUTTON 11
#define BUZZER 10

#define NO_BLOCKS 7
#define BLOCK_SIZE 50
#define HEIGHT 480
#define WIDTH 320
#define PAD_UP 19
#define PAD_LEFT 10
#define PAD_RIGHT 10
#define GRAVITY 20
#define TEXT_OFFSET 70

#define PAD_UP_SCORE 3
#define PAD_RIGHT_SCORE 150


#define TOP_LIMIT 160

#define ACC_THRESHOLD 3
#define MOVE_SPEED 20


enum game_state {
  RUNNING,
  GAME_OVER
};

game_state g_state;

enum state {
  MOVE,
  FALL
};

struct color_rgb {
  uint8_t r, g, b;
};

struct block {
  int x;
  int y;
  color_rgb color;
};

struct speed {
  int8_t v_x;
  int8_t v_y;
};

const color_rgb RED = {255, 0 , 0};
const color_rgb GREEN = {0, 255, 0};
const color_rgb BLUE = {0, 0, 255};
const color_rgb BLACK = {0, 0, 0};
const color_rgb MAGENTA = {255, 0, 255};
const color_rgb YELLOW = {255, 255, 0};
const color_rgb ORANGE = {255, 165, 0};
const color_rgb SALMON = {255, 160, 122};
const color_rgb SPRING_GREEN = {0, 250, 154};
const color_rgb AQUA_MARINE = {102, 205, 170};
const color_rgb CYAN = {0, 255, 255};
const color_rgb PINK = {255, 105, 180};

state curr_state;
block curr_block;
speed curr_speed;

uint8_t score;

block *top;

block v_block[NO_BLOCKS];
uint8_t no_blocks;


void setup_sensor() {
  mpu.begin();

  mpu.setAccelerometerRange(MPU6050_RANGE_8_G);
  mpu.setGyroRange(MPU6050_RANGE_500_DEG);
  mpu.setFilterBandwidth(MPU6050_BAND_21_HZ);
}
void add_block(block *b) {
  // add to vector
  v_block[no_blocks].y = b->y;
  v_block[no_blocks].x = b->x;
  v_block[no_blocks].color = b->color;

  // draw block
  draw_block(b);

  // set new top
  top = v_block + no_blocks;

  // increase vector size
  no_blocks++;

  score++;
  print_score();

  if (top->y <= TOP_LIMIT) {
    shift_blocks_down();
  }
}

void init_game() {
  g_state = RUNNING;
  clear_screen();

  no_blocks = 0;
  score = -1;
  // initialize first block
  block tmp = {WIDTH / 2 - BLOCK_SIZE / 2, HEIGHT - BLOCK_SIZE, RED};
  add_block(&tmp);

  // initialize current block
  new_current_block();


}
ISR(PCINT0_vect) {
  if (g_state == GAME_OVER && ((PINB & (1 << PINB3)) == 0)) {
    init_game();
  }
  // if current state is moving and the button is pressed
  if (curr_state == MOVE && ((PINB & (1 << PINB3)) == 0)) {
    // make the block fall
    curr_state = FALL;
    curr_speed.v_x = 0;
    curr_speed.v_y = GRAVITY;
  }
}
color_rgb random_color() {
  static color_rgb vec[] = {RED, GREEN, BLUE, MAGENTA, YELLOW, ORANGE,
    SALMON, SPRING_GREEN, AQUA_MARINE, CYAN, PINK};
  static uint8_t size = sizeof(vec) / sizeof(color_rgb);

  uint8_t idx = rand() % size;

  return vec[idx];
}
void clear_screen() {
  myGLCD.setColor(0, 0, 0);
  myGLCD.fillRect(0, 0, WIDTH, HEIGHT);
}

void draw_block(block *b) {
  myGLCD.setColor(b->color.r, b->color.g, b->color.b);
  myGLCD.fillRect(b->x, b->y, b->x + BLOCK_SIZE, b->y + BLOCK_SIZE);
}

void new_current_block() {
  curr_block = {rand() % (WIDTH - BLOCK_SIZE) , PAD_UP, random_color()};
  draw_block(&curr_block);

  // set initial speed
  curr_speed = {(rand() % 2 == 0 ? 1 : -1) * 20, 0};

  // set state
  curr_state = MOVE;
}

void print_score() {
  char str_score[30];

  sprintf(str_score, "%s: %d", "SCORE", score);
  myGLCD.print(str_score, WIDTH - PAD_RIGHT_SCORE, PAD_UP_SCORE);
}

void shift_blocks_down() {
  color_rgb tmp;
  for (int i = 0; i < no_blocks; ++i) {
    block *b = v_block + i;

    tmp = b->color;

    // set old block black
    b->color = BLACK;
    draw_block(b);

    // update new position
    b->y += 3 * BLOCK_SIZE;

    // draw new block
    b->color = tmp;
    draw_block(b);
  }

  // remove blocks and shift the vector
  for (int i = 0; i < no_blocks - 3; ++i) {
    v_block[i] = v_block[i + 3];
  }

  no_blocks -= 3;

  top = v_block + no_blocks - 1;
}

void setup()
{
  // Setup the LCD
  myGLCD.InitLCD(PORTRAIT);
  myGLCD.setFont(BigFont);

  // configure button interrupt
  
  // enable PCINT0
  PCICR = (1 << PCIE0);
  // enable interrupt on pin PCINT3
  PCMSK0 = (1 << PCINT3);

  setup_sensor();

  init_game();
}

void sound_game_over() {
  static int melody[] = {
    NOTE_C4, NOTE_G3, NOTE_G3, NOTE_A3, NOTE_G3, 0, NOTE_B3, NOTE_C4
  };

  // note durations: 4 = quarter note, 8 = eighth note, etc.:
  static int noteDurations[] = {
    4, 8, 8, 4, 4, 4, 4, 4
  };
  for (int thisNote = 0; thisNote < 8; thisNote++) {

    // to calculate the note duration, take one second divided by the note type.
    //e.g. quarter note = 1000 / 4, eighth note = 1000/8, etc.
    int noteDuration = 1000 / noteDurations[thisNote];
    tone(10, melody[thisNote], noteDuration);

    // to distinguish the notes, set a minimum time between them.
    // the note's duration + 30% seems to work well:
    int pauseBetweenNotes = noteDuration * 1.30;
    delay(pauseBetweenNotes);
    // stop the tone playing:
    noTone(8);
  }
}

void game_over() {
  g_state = GAME_OVER;

  clear_screen();

  myGLCD.setColor(255, 0, 0);
  myGLCD.print("GAME OVER", WIDTH / 2 - TEXT_OFFSET, HEIGHT / 2);
  sound_game_over();
}

void loop()
{

  if (g_state == GAME_OVER) {
    return;
  }

  color_rgb old_color = curr_block.color;

  // clear the old block position
  curr_block.color = BLACK;
  draw_block(&curr_block);

  // reset color back to the original one
  curr_block.color = old_color;

  // if the block is moving left-right, make sure it doesn't go out of bounds
  if (curr_state == MOVE) {
    sensors_event_t a, g, temp;
    mpu.getEvent(&a, &g, &temp);
    if (a.acceleration.x <= -ACC_THRESHOLD) {
      curr_speed.v_x = MOVE_SPEED;
    }
  
    if (a.acceleration.x >= ACC_THRESHOLD) {
      curr_speed.v_x = -MOVE_SPEED;
    }
  
    if (a.acceleration.x >= -ACC_THRESHOLD && a.acceleration.x <= ACC_THRESHOLD) {
      curr_speed.v_x = 0;
    }
  }

  curr_block.x += curr_speed.v_x;
  curr_block.y += curr_speed.v_y;

  if (curr_block.x <= 0) {
    curr_block.x = 0;
  }

  if (curr_block.x >= WIDTH - BLOCK_SIZE) {
    curr_block.x = WIDTH - BLOCK_SIZE;
  }
  // check collisions with tower
  if (curr_state == FALL) {
    int lower_y = curr_block.y + BLOCK_SIZE;
    int left_x = curr_block.x + BLOCK_SIZE;
    // if I should check for collisions
    if (lower_y >= top->y) {
      // check if the block can land on top of the tower
      if (curr_block.x >= top->x && curr_block.x <= top->x + BLOCK_SIZE
          || left_x >= top->x && left_x <= top->x + BLOCK_SIZE) {
        // place the block
        curr_block.y = top->y - BLOCK_SIZE;
        add_block(&curr_block);

        new_current_block();

        return;
      }
    }
  }

  if (curr_state == FALL && HEIGHT - curr_block.y <= BLOCK_SIZE) {
    draw_block(&curr_block);
    game_over();
    return;
  }

  draw_block(&curr_block);

  delay(250);
}
